use nom::branch::alt;
use nom::bytes::complete::{tag, take_while1};
use nom::character::complete::{anychar, char as take_char};
use nom::combinator::map;
use nom::multi::{many0, many_till};
use nom::sequence::tuple;
use nom::IResult;

pub(crate) fn parse_till_mul(input: &str) -> IResult<&str, usize> {
    map(many_till(anychar, parse_mul), |(_, product)| product)(input)
}

pub(crate) fn parse_mul(input: &str) -> IResult<&str, usize> {
    map(
        tuple((
            tag::<_, &str, _>("mul("),
            take_while1(|c: char| c.is_ascii_digit()),
            take_char(','),
            take_while1(|c: char| c.is_ascii_digit()),
            take_char(')'),
        )),
        |(_, first, _, second, _)| {
            if let (Ok(first), Ok(second)) = (first.parse::<usize>(), second.parse::<usize>()) {
                return first * second;
            }
            panic!("Failed to parse")
        },
    )(input)
}

pub(crate) fn parse_input(input: &str) -> IResult<&str, Vec<usize>> {
    many0(parse_till_mul)(input)
}

pub(crate) fn parse_do(input: &str) -> IResult<&str, ()> {
    map(tag("do()"), |_| ())(input)
}

pub(crate) fn parse_dont(input: &str) -> IResult<&str, ()> {
    map(tag("don't()"), |_| ())(input)
}

#[derive(Default)]
pub(crate) struct Parser {
    do_mul: bool,
    sum: usize,
}

impl Parser {
    pub(crate) const fn new() -> Self {
        Self {
            do_mul: true,
            sum: 0,
        }
    }

    pub(crate) fn parse<'a>(&mut self, input: &'a str) -> IResult<&'a str, ()> {
        map(many0(|input| self.parse_till_expression(input)), |_| ())(input)
    }

    pub(crate) const fn sum(&self) -> &usize {
        &self.sum
    }

    fn parse_till_expression<'a>(&mut self, input: &'a str) -> IResult<&'a str, ()> {
        let (remaining, (_, (sum, do_mul))) = many_till(
            anychar,
            alt((
                map(parse_dont, |_| (0, false)),
                map(parse_do, |_| (0, true)),
                map(parse_mul, |sum| {
                    if self.do_mul {
                        (sum, true)
                    } else {
                        (0, false)
                    }
                }),
            )),
        )(input)?;

        self.sum += sum;
        self.do_mul = do_mul;

        Ok((remaining, ()))
    }
}
